"""Calculate expected CLs values with hypothesis tests."""
from __future__ import annotations

__all__ = ("hypotest",)

from functools import partial

import jax
import jax.numpy as jnp
import pyhf
from chex import Array

from ..mle import fit, fixed_poi_fit


@partial(jax.jit, static_argnames=["model", "return_mle_pars"])  # forward pass
def hypotest(
    test_poi: float,
    data: Array,
    model: pyhf.Model,
    lr: float,
    return_mle_pars: bool = False,
) -> tuple[Array, Array] | Array:
    # hard-code 1 as inits for now
    # TODO: need to parse different inits for constrained and global fits
    init_pars = jnp.ones_like(jnp.asarray(model.config.suggested_init()))
    conditional_pars = fixed_poi_fit(
        data, model, poi_condition=test_poi, init_pars=init_pars[:-1], lr=lr
    )
    mle_pars = fit(data, model, init_pars=init_pars, lr=lr)
    profile_likelihood = -2 * (
        model.logpdf(conditional_pars, data)[0] - model.logpdf(mle_pars, data)[0]
    )

    poi_hat = mle_pars[model.config.poi_index]
    qmu = jnp.where(poi_hat < test_poi, profile_likelihood, 0.0)

    CLsb = 1 - pyhf.tensorlib.normal_cdf(jnp.sqrt(qmu))
    altval = 0.0
    CLb = 1 - pyhf.tensorlib.normal_cdf(altval)
    CLs = CLsb / CLb
    return (CLs, mle_pars) if return_mle_pars else CLs



from __future__ import annotations

__all__ = ("fit",)

from functools import partial
from typing import TYPE_CHECKING, Callable, cast

import jax
from chex import Array

if TYPE_CHECKING:
    import pyhf

from .minimize import _minimize


def global_fit_objective(data: Array, model: pyhf.Model) -> Callable[[Array], float]:
    def fit_objective(lhood_pars_to_optimize: Array) -> float:  # NLL
        """lhood_pars_to_optimize: either all pars, or just nuisance pars"""
        return cast(
            float, -model.logpdf(lhood_pars_to_optimize, data)[0]
        )  # pyhf.Model.logpdf returns list[float]

    return fit_objective


@partial(jax.jit, static_argnames=["model"])
def fit(
    data: Array,
    model: pyhf.Model,
    init_pars: Array,
    lr: float = 1e-2,
) -> Array:
    obj = global_fit_objective(data, model)
    fit_res = _minimize(obj, init_pars, lr)
    return fit_res



import jax.numpy as jnp
import pyhf


# class-based
class _Config:
    def __init__(self):
        self.poi_index = 0
        self.npars = 2

    def suggested_init(self):
        return jnp.asarray([1.0, 1.0])

    def suggested_bounds(self):
        return jnp.asarray([jnp.asarray([0.0, 10.0]), jnp.asarray([0.0, 10.0])])


class Model:
    """Dummy class to mimic the functionality of `pyhf.Model`."""

    def __init__(self, spec):
        pyhf.set_backend("jax")
        self.sig, self.nominal, self.uncert = spec
        self.factor = (self.nominal / self.uncert) ** 2
        self.aux = 1.0 * self.factor
        self.config = _Config()

    def expected_data(self, pars):
        mu, gamma = pars
        expected_main = jnp.asarray([gamma * self.nominal + mu * self.sig])
        aux_data = jnp.asarray([self.aux])
        return jnp.concatenate([expected_main, aux_data])

    def logpdf(self, pars, data):

        maindata, auxdata = data
        main, _ = self.expected_data(pars)
        _, gamma = pars
        main = pyhf.probability.Poisson(main).log_prob(maindata)
        constraint = pyhf.probability.Poisson(gamma * self.factor).log_prob(auxdata)
        # sum log probs over bins
        return [jnp.sum(jnp.asarray([main + constraint]), axis=None)]


def uncorrelated_background(signal_data, bkg_data, bkg_uncerts):
    """Dummy class to mimic the functionality of `pyhf.simplemodels.hepdata_like`."""
    return Model([signal_data, bkg_data, bkg_uncerts])



import jax.numpy as jnp
import pytest
from dummy_pyhf import uncorrelated_background


@pytest.fixture
def example_model():
    return uncorrelated_background(
        signal_data=jnp.asarray([5, 5]),
        bkg_data=jnp.asarray([50, 50]),
        bkg_uncerts=jnp.asarray([5, 5]),
    )



from __future__ import annotations

__all__ = (
    "fisher_info",
    "cramer_rao_uncert",
)

from typing import Callable

import jax
import jax.numpy as jnp
from chex import Array


def fisher_info(
    logpdf: Callable[[Array, Array], float], pars: Array, data: Array
) -> Array:
    return -jax.hessian(logpdf)(pars, data)


def cramer_rao_uncert(
    logpdf: Callable[[Array, Array], float], pars: Array, data: Array
) -> Array:
    inv = jnp.linalg.inv(fisher_info(logpdf, pars, data))
    return jnp.sqrt(jnp.diagonal(inv))



from __future__ import annotations

__all__ = ("fixed_poi_fit",)

from functools import partial
from typing import TYPE_CHECKING, Callable, cast

import jax
import jax.numpy as jnp
from chex import Array

if TYPE_CHECKING:
    import pyhf

from .minimize import _minimize


def fixed_poi_fit_objective(
    data: Array,
    model: pyhf.Model,
) -> Callable[[Array, float], float]:
    poi_idx = model.config.poi_index

    def fit_objective(
        lhood_pars_to_optimize: Array, poi_condition: float
    ) -> float:  # NLL
        """lhood_pars_to_optimize: either all pars, or just nuisance pars"""
        # pyhf.Model.logpdf returns list[float]
        blank = jnp.zeros_like(jnp.asarray(model.config.suggested_init()))
        blank += lhood_pars_to_optimize
        return cast(float, -model.logpdf(blank.at[poi_idx].set(poi_condition), data)[0])

    return fit_objective


@partial(jax.jit, static_argnames=["model"])  # forward pass
def fixed_poi_fit(
    data: Array,
    model: pyhf.Model,
    init_pars: Array,
    poi_condition: float,
    lr: float = 1e-2,  # arbitrary
) -> Array:
    obj = fixed_poi_fit_objective(data, model)
    fit_res = _minimize(obj, init_pars, lr, poi_condition)
    blank = jnp.zeros_like(jnp.asarray(model.config.suggested_init()))
    blank += fit_res
    poi_idx = model.config.poi_index
    return blank.at[poi_idx].set(poi_condition)



import jax.numpy as jnp
import numpy as np
import pyhf
from dummy_pyhf import uncorrelated_background
from jax import jacrev

import relaxed


def test_fit(example_model):
    analytic_pars = jnp.array([0.0, 1.0])

    mle_pars = relaxed.mle.fit(
        model=example_model,
        data=example_model.expected_data(analytic_pars),
        init_pars=example_model.config.suggested_init(),
        lr=1e-2,
    )
    assert np.allclose(mle_pars, analytic_pars, atol=0.01)


def test_fit_grad():
    def pipeline(x):
        model = uncorrelated_background(x * 5.0, x * 20, x * 2)
        mle_pars = relaxed.mle.fit(
            model=model,
            data=model.expected_data(jnp.array([0.0, 1.0])),
            init_pars=model.config.suggested_init(),
            lr=1e-2,
        )
        return mle_pars

    jacrev(pipeline)(jnp.asarray(0.5))


def test_fixed_poi_fit():
    pyhf.set_backend("jax")
    analytic_pars = jnp.array([0.0, 1.0])

    example_model = uncorrelated_background(
        signal_data=jnp.asarray([5]),
        bkg_data=jnp.asarray([50]),
        bkg_uncerts=jnp.asarray([5]),
    )
    init = np.asarray(example_model.config.suggested_init())
    init = jnp.asarray(np.delete(init, example_model.config.poi_index))
    relaxed_mle = relaxed.mle.fixed_poi_fit(
        model=example_model,
        data=example_model.expected_data(analytic_pars),
        init_pars=init,
        lr=1e-2,
        poi_condition=1.0,
    )

    m = pyhf.simplemodels.uncorrelated_background([5], [50], [5])

    pyhf_mle = pyhf.infer.mle.fixed_poi_fit(
        1.0,
        m.expected_data(analytic_pars),
        m,
    )

    assert np.allclose(relaxed_mle, pyhf_mle, rtol=1e-4)


def test_fixed_poi_fit_grad():
    def pipeline(x):
        model = uncorrelated_background(x * 5.0, x * 20, x * 2)
        mle_pars = relaxed.mle.fixed_poi_fit(
            model=model,
            data=model.expected_data(jnp.array([0.0, 1.0])),
            init_pars=model.config.suggested_init()[1:],
            lr=1e-2,
            poi_condition=1.0,
        )
        return mle_pars

    jacrev(pipeline)(jnp.asarray(0.5))



"""differentiable implementation of the histogram via kernel density estimation."""
from __future__ import annotations

__all__ = ("hist",)

from functools import partial

import jax
import jax.numpy as jnp
import jax.scipy as jsp
from chex import Array


@partial(jax.jit, static_argnames=["density", "reflect_infinities"])
def hist(
    events: Array,
    bins: Array,
    bandwidth: float,  # | None = None,
    density: bool = False,
    reflect_infinities: bool = False,
) -> Array:
    """
    Differentiable implementation of a histogram using kernel density estimation.

    Parameters
    ----------
    events: (jax array-like) 1D data!
    bins: (jax array-like) intervals to calculate counts.
    bandwidth: (float) value that specifies the width of the individual
        distributions (kernels), whose cdfs are averaged over each bin. Defaults
        to Scott's rule -- the same as scipy's.
    density: (bool) whether or not to normalize the histogram to unit area.
    reflect_infinities: (bool) if true, reflect  under/overflow bins into boundary bins.
    doing so will ensure (normalised) unit total density,
    as kdes have infinite support.

    Returns
    -------
    counts: 1D array of binned counts
    """
    # bandwidth = bandwidth or events.shape[-1] ** -0.25  # Scott's rule

    bins = jnp.array([-jnp.inf, *bins, jnp.inf]) if reflect_infinities else bins

    edge_hi = bins[1:]  # ending bin edges ||<-
    edge_lo = bins[:-1]  # starting bin edges ->||

    # get cumulative counts (area under kde) for each set of bin edges
    cdf_up = jsp.stats.norm.cdf(edge_hi.reshape(-1, 1), loc=events, scale=bandwidth)
    cdf_dn = jsp.stats.norm.cdf(edge_lo.reshape(-1, 1), loc=events, scale=bandwidth)
    # sum kde contributions in each bin
    counts = (cdf_up - cdf_dn).sum(axis=1)

    if density:  # normalize by bin width and counts for total area = 1
        db = jnp.array(jnp.diff(bins), float)  # bin spacing
        counts = counts / db / counts.sum(axis=0)

    if reflect_infinities:
        counts = (
            counts[1:-1]
            + jnp.array([counts[0]] + [0] * (len(counts) - 3))
            + jnp.array([0] * (len(counts) - 3) + [counts[-1]])
        )

    return counts



from functools import partial

import jax.numpy as jnp
import numpy as np
import pyhf
import pytest
from jax import jacrev, vmap
from jax.random import PRNGKey, normal

import relaxed


@pytest.fixture
def big_sample():
    return normal(PRNGKey(0), shape=(5000,))


@pytest.fixture
def bins():
    return np.linspace(-5, 5, 6)


def test_hist_validity(big_sample, bins):
    numpy_hist = np.histogram(big_sample, bins=bins)[0]
    relaxed_hist = relaxed.hist(big_sample, bins=bins, bandwidth=1e-6)
    assert np.allclose(numpy_hist, relaxed_hist)


def test_hist_validity_density(big_sample, bins):
    numpy_hist = np.histogram(big_sample, bins=bins, density=True)[0]
    relaxed_hist = relaxed.hist(big_sample, bins=bins, bandwidth=1e-6, density=True)
    assert np.allclose(numpy_hist, relaxed_hist)


def test_hist_validity_infinities(big_sample, bins):
    """Test the reflection of excess density @ [-inf,inf] into the edge bins."""
    inf_bins = [-np.inf, *bins, np.inf]
    numpy_hist = np.histogram(big_sample, bins=inf_bins, density=True)[0]
    numpy_hist[1] += numpy_hist[0]
    numpy_hist[-2] += numpy_hist[-1]
    numpy_hist = numpy_hist[1:-1]
    relaxed_hist = relaxed.hist(
        big_sample, bins=bins, bandwidth=1e-6, density=True, reflect_infinities=True
    )
    assert np.allclose(numpy_hist, relaxed_hist)


def test_hist_approx_validity(big_sample, bins):
    """Roughly test validity of hist for wider bandwidth (and a more loose criterion).
    Useful because it's the same bandwidth (0.15) as the gradient test below."""
    numpy_hist = np.histogram(big_sample, bins=bins, density=True)[0]
    relaxed_hist = relaxed.hist(big_sample, bins=bins, bandwidth=0.15, density=True)
    assert np.allclose(numpy_hist, relaxed_hist, atol=0.01)


def test_hist_grad_validity(bins):
    """Test the grads of the kde hist vs the analyitc grads of a normal dist wrt mu."""

    def gen_points(mu, jrng, nsamples):
        points = normal(jrng, shape=(nsamples,)) + mu
        return points

    def bin_height(mu, jrng, bw, nsamples, bins):
        points = gen_points(mu, jrng, nsamples)
        return relaxed.hist(points, bins, bandwidth=bw)

    mus = jnp.linspace(-2, 2, 100)

    def kde_grads(bw, nsamples):
        rngs = [PRNGKey(i) for i in range(5)]
        grad_fun = jacrev(bin_height)
        grads = []
        for jrng in rngs:
            get_grads = vmap(
                partial(grad_fun, jrng=jrng, bw=bw, nsamples=nsamples, bins=bins)
            )
            grads.append(get_grads(mus))
        return jnp.asarray(grads)

    nsamples = 5000
    relaxed_grads = kde_grads(bw=0.15, nsamples=nsamples).mean(axis=0) / nsamples

    def true_grad(mu, bins):
        """Analytic grad of the mean height over an interval of a normal dist wrt mu."""
        # The full equation in latex for copy/paste (delete the # statement at the end):
        # \frac{\partial}{\partial\mu}bin_{\mathsf{true}}(\mu) = -\frac{1}{\sqrt{2\\pi}}\left[\left(e^{-\frac{(b-\mu)^2}{2}}\right) - \left( e^{-\frac{(a-\mu)^2}{2}}\right)\right] # fmt: skip
        b = bins[1:]  # ending bin edges ||<-
        a = bins[:-1]  # starting bin edges ->||
        return -(1 / ((2 * jnp.pi) ** 0.5)) * (
            jnp.exp(-((b - mu) ** 2) / 2) - jnp.exp(-((a - mu) ** 2) / 2)
        )

    grads = vmap(partial(true_grad, bins=bins))(mus)

    assert np.allclose(
        relaxed_grads, grads, atol=0.01, rtol=0.05
    )  # tols are a bit high because the grads are a little noisy/biased


def test_fisher_info(example_model):
    def model(pars, data):
        return example_model.logpdf(pars, data)[0]

    pars = example_model.config.suggested_init()
    data = example_model.expected_data(pars)

    # this is just the computed output, assumed correct
    # probably needs a more thorough analytic test
    res = np.array([[0.90909091, 9.09090909], [9.09090909, 290.90909091]])

    assert np.allclose(relaxed.fisher_info(model, pars, data), res)


def test_fisher_uncerts_validity():
    pyhf.set_backend("jax", pyhf.optimize.minuit_optimizer(verbose=1))

    m = pyhf.simplemodels.uncorrelated_background([5, 5], [50, 50], [5, 5])

    data = jnp.array([50.0, 50.0] + m.config.auxdata)

    fit_res = pyhf.infer.mle.fit(
        data,
        m,
        return_uncertainties=True,
        par_bounds=[
            [-1, 10],
            [-1, 10],
            [-1, 10],
        ],  # fit @ boundary produces unstable uncerts
    )

    # minuit fit uncerts
    mle_pars, mle_uncerts = fit_res[:, 0], fit_res[:, 1]

    # uncertainties from autodiff hessian
    def lpdf(p, d):
        return m.logpdf(p, d)[0]

    relaxed_uncerts = relaxed.cramer_rao_uncert(lpdf, mle_pars, data)
    assert np.allclose(mle_uncerts, relaxed_uncerts, rtol=5e-2)


def test_fisher_info_grad(example_model):
    def pipeline(x):
        pars = example_model.config.suggested_init()

        def model(pars, data):
            return example_model.logpdf(pars, data)[0]

        data = example_model.expected_data(pars)
        return relaxed.fisher_info(model, pars * x, data * x)

    jacrev(pipeline)(4.0)  # just check you can calc it w/o exception


def test_fisher_uncert_grad(example_model):
    def pipeline(x):
        def model(pars, data):
            return example_model.logpdf(pars, data)[0]

        pars = example_model.config.suggested_init()
        data = example_model.expected_data(pars)
        return relaxed.cramer_rao_uncert(model, pars * x, data * x)

    jacrev(pipeline)(4.0)  # just check you can calc it w/o exception



from relaxed._version import version as __version__

__all__ = ("__version__", "hist", "cramer_rao_uncert", "fisher_info", "mle", "infer")

from relaxed import infer, mle
from relaxed.ops import cramer_rao_uncert, fisher_info, hist



import jax.numpy as jnp
import numpy as np
import pyhf
from dummy_pyhf import uncorrelated_background
from jax import jacrev

import relaxed


def test_hypotest_validity(example_model):
    pyhf.set_backend("jax")
    analytic_pars = jnp.array([0.0, 1.0])
    relaxed_cls = relaxed.infer.hypotest(
        1, example_model.expected_data(analytic_pars), example_model, lr=1e-2
    )
    m = pyhf.simplemodels.uncorrelated_background([5], [50], [5])
    pyhf_cls = pyhf.infer.hypotest(1, m.expected_data(analytic_pars), m)
    np.allclose(relaxed_cls, pyhf_cls)


def test_hypotest_grad():
    def pipeline(x):
        model = uncorrelated_background(x * 5.0, x * 20, x * 2)
        expected_cls = relaxed.infer.hypotest(
            1.0,
            model=model,
            data=model.expected_data(jnp.array([0.0, 1.0])),
            lr=1e-2,
        )
        return expected_cls

    jacrev(pipeline)(jnp.asarray(0.5))



from __future__ import annotations

from typing import Any, Callable

__all__ = ("_minimize",)

from functools import partial

import jax
import jaxopt
import optax
from chex import Array


# try wrapping obj with closure_convert
@partial(jax.jit, static_argnames=["objective_fn"])  # forward pass
def _minimize(
    objective_fn: Callable[..., float], init_pars: Array, lr: float, *obj_args: Any
) -> Array:
    converted_fn, aux_pars = jax.closure_convert(objective_fn, init_pars, *obj_args)
    # aux_pars seems to be empty? took that line from jax docs example...
    solver = jaxopt.OptaxSolver(
        fun=converted_fn, opt=optax.adam(lr), implicit_diff=True, maxiter=5000
    )
    return solver.run(init_pars, *obj_args, *aux_pars)[0]
